﻿B4J=true
Group=Default Group
ModulesStructureVersion=1
Type=Class
Version=8.5
@EndOfDesignText@
'Introduction:

'This class presents a page that contains both code and results from the code
'This is particularly useful to show how novel processes work, step by step.
'A somewhat similar effect can be achieved by putting breakpoints in the code, and using F9 in debug mode.
'I found that the F9 method is fine for debugging but not ideal for demonstrating code examples

'Here are some advantages to using WiLDemo:
'	1. the user does not have to manage the debug/breakpoint part (or look at the code at all)
'	2. the steps are fully contained on a B4Xpage (if using B4XPages)
'	3. it works in release and debug mode in B4J and B4A
'	4. changes made to the code show up in the demo (after saving/copying the code  using the link below)
'	5. the examples are named and managed naturally

'The method is achieved by having a copy of B4XMainPAge in the Shared Files folder
'When an instance of the class is initialized, that copy of the B4XMainPAge is read from the file
'The lines between "demo.begin" and "Wait For demo_done" are identified as wel as the Sub name wehere they are
'When begin is run the first time, the page is shown with the code of first example, and then returns to MainPAge
'where the code is run, and the process waits for the user to to click the next button.

'The next button generates a "demo_done" event, with the instruction callSubDelayed(mCallBack, "demo_done")

'If you change B4XMainPage then save program first and then Ctrl + click on the following to copy B4XMainPage.bas to Shared Files folder:
'			 ide://run?file=%WINDIR%\System32\Robocopy.exe&args=..\..\&args=..\..\Shared+Files&args=B4XMainPage.bas

'You can also disable the demo with "demo.Initialize(Me, False)" - The begin Sub will then do nothing except generate the demo_done event
'You don't need to remove the "demo.begin" or "Wait For demo_done", unless you don't want them anymore

'The results from the execution of the code are sent to the demo page
'The "display" sub redirects the results to myLog in calling module


Sub Class_Globals
	Private Root As B4XView 'ignore
	Private xui As XUI 'ignore
	Private mCallBack As Object

	Private textEngine As BCTextEngine
	Private demonstration As BBCodeView
	Private Button1 As B4XView
	Private demoCode As List
	Private demoMap As Map
	Private contentsSofar As String
	Public enabled As Boolean
	Public inDemoMode As Boolean
	Private lastSubName As String
	Private lastDemoName As String
	Public isLastDemo As Boolean
	
	Private MonoFont As B4XFont
	Private FontSize As Int = 18
	Private more As Label
End Sub

Public Sub Initialize(callBack As Object, moduleName As String, enabled1 As Boolean) As Object
	mCallBack = callBack
	enabled = enabled1
	demoMap.initialize
	inDemoMode = False
	If enabled Then
		demoCode = File.ReadList(File.DirAssets, moduleName & ".bas")
		Dim subName As String = ""
		For i = 0 To demoCode.Size - 1
			Dim s As String = demoCode.get(i)
			s = s.trim
			If s.Contains("Sub ") And s.StartsWith("'") = False Then
				If s.Contains(TAB & "'") Then s = s.SubString2(0, s.IndexOf2("'", 4)) 
				If s.contains("(") Then
					subName = s.SubString2(s.IndexOf("Sub") + 3, s.IndexOf("(")).trim
				Else
					subName = s.SubString(s.IndexOf("Sub") + 3).trim
				End If
			End If
			If s.StartsWith("demo.begin(") Then 
				Dim DemoName As String = s.SubString2(12, s.IndexOf2(QUOTE, 12))
				If subName <> lastSubName And lastSubName <> "" Then 
					Dim lastref As String = demoMap.Get(lastDemoName)
					demoMap.Put(lastDemoName, "*" & lastref)
				End If
				demoMap.Put(DemoName, i & TAB & subName & ": " & DemoName)
				lastSubName = subName
				lastDemoName = DemoName
			End If
		Next
		Dim lastref As String = demoMap.Get(lastDemoName)
		demoMap.Put(lastDemoName, "*" & lastref)

		#if B4J
			Dim fx As JFX
			Dim monospace As Font = fx.CreateFont("Consolas", FontSize, False, False)
		#Else If B4A
			Dim monospace As Typeface = Typeface.CreateNew(Typeface.MONOSPACE, Typeface.STYLE_NORMAL)
		#Else If B4i
			Dim monospace As Font = Font.CreateNew2("DevanagariSangamMN", FontSize)
		#End If
		MonoFont = xui.CreateFont(monospace, FontSize)
	End If
	Return Me
End Sub

Private Sub B4XPage_Created (Root1 As B4XView)
	Root = Root1
	Root.LoadLayout("1")
	Root.Color = xui.color_RGB(210, 210, 210)

	textEngine.Initialize(Root)
	textEngine.KerningEnabled = False
	textEngine.CustomFonts.Put("monospace", MonoFont)
	demonstration.TextEngine = textEngine
End Sub

Public Sub begin(s As String)
	If enabled = False Then
		CallSubDelayed(mCallBack, "demo_done")
		Return
	End If
	inDemoMode = True
	B4XPages.showPage("demo")
	Dim codeRef As String = demoMap.get(s)
	isLastDemo = False
	If codeRef.StartsWith("*") Then 
		codeRef = codeRef.SubString(1)
		isLastDemo = True
	End If
	Dim k As Int = codeRef.IndexOf(TAB)
	Dim startIndex As Int = codeRef.substring2(0, k)
	Dim title As String = codeRef.substring(k+1)
	demonstration.sv.ScrollViewOffsetY = 0
	Dim content As String = _	
$"[font=monospace]
[Alignment=Center][TextSize=20][b]${title}[/b][/TextSize][/Alignment]
[TextSize=18]
"$
	Dim sb As StringBuilder: sb.Initialize
	sb.Append(content)
	For j = startIndex + 1 To demoCode.Size - 1
		Dim s As String = demoCode.get(j)
		If s.startsWith(TAB) Then s = s.SubString(1)
		If s.contains("demo_done") Then Exit
		If s.StartsWith("'") Then s = "[Color=#008800]   " & s & "[/Color]"
		sb.append(s).Append(CRLF)
	Next
	sb.append(TAB).Append(CRLF)
	contentsSofar = sb.toString
	demonstration.Text = contentsSofar
	
	'Code Snippet Alert - check if the BBCodeview ccontents need to be scrolled
	If demonstration.Paragraph.Height / textEngine.mScale + demonstration.Padding.Top + demonstration.Padding.Bottom > demonstration.mBase.height Then
		more.visible = True
	Else
		more.visible = False
	End If
End Sub

Public Sub addLogs(s As String, info As String)
	Dim sb As StringBuilder:sb.Initialize
	sb.Append($"_________ ${info} _________"$).Append(CRLF)
	If s.Length > 0 Then sb.append(s).append(CRLF)
	contentsSofar = contentsSofar & sb.toString
	demonstration.Text = contentsSofar
	Dim ph As Float = demonstration.Paragraph.Height - 40dip
	If ph / textEngine.mScale + demonstration.Padding.Top + demonstration.Padding.Bottom > demonstration.mBase.height Then 
		more.visible = True
	Else
		more.visible = False
	End If
End Sub

Private Sub Button1_Click
	contentsSofar = ""
	inDemoMode = False
	lastDemoName = ""
	
	'Code Snippet Alert - generates a Wait For event in the mCallBack context
	CallSubDelayed(mCallBack, "demo_done")
	If isLastDemo Then
		B4XPages.ClosePage(Me)
		isLastDemo = False
	End If
End Sub

'this matrix error handler in B4XMainPage is essential to the dMatrix class
Public Sub matrixErrors(s As String)
	If s.StartsWith("?") Then
		myLog(CRLF & s.substring(1) & CRLF & CRLF & "Calling Sequence: " & CRLF & getStackTrace("errorMessage"), "WARNING")
	Else If s.StartsWith("!") Then 
		myLog(CRLF & s.substring(1) & CRLF & CRLF & "Calling Sequence: " & CRLF & getStackTrace("errorMessage"), "ALERT")
	Else
		myLog(CRLF & s & CRLF & CRLF & "Calling Sequence: " & getStackTrace("errorMessage") & CRLF & "EXECUTION STOPPED", "ERROR")
		ExitApplication
	End If
End Sub

'for displaying one-dimensional arrays
Public Sub vectorToString (z() As Double, nDecimal As Int) As String
	Dim sb As StringBuilder: sb.initialize
	For i = 0 To z.Length - 1
		sb.Append(NumberFormat2(z(i), 1, nDecimal, nDecimal, False)).append(TAB)
	Next
	Return sb.tostring
End Sub

'if demo is enabled, log output is passed to demo page
Public Sub myLog(s As String, info As String)
	Dim logResult As Boolean = True
	Dim allPages As B4XSet = B4XPages.GetManager.mStackOfPageIds
	If allPages.Contains("demo") Then
		Dim dp As WiLDemo = B4XPages.GetPage("demo")
		If dp.inDemoMode Then 
			If info="WARNING" Or info="ALERT" Or info="ERROR" Then 
				dp.addLogs(s, $"[Plain]${info}[/Plain]"$)
			Else
				dp.addLogs(toBBC(s), $"[Plain]${info}[/Plain]"$)
			End If
			logResult = False
		End If
	End If
	If logResult Then
		Log($"________  ${info}  ________"$)
		Log(s)
		Log(TAB)
	End If
End Sub

'Thanks to SteveI05 https://www.b4x.com/android/forum/threads/log-stack-trace.62979/
Public Sub getStackTrace(startAtSub As String) As String
	Dim result As List: result.Initialize
	Dim StackTrace() As Object
	Dim Throwable As JavaObject
	Dim i As Int = 0
	Dim relevantSection As Boolean = False
	Dim JO As JavaObject
	Throwable.InitializeNewInstance("java.lang.Throwable",Null)
	StackTrace = Throwable.RunMethod("getStackTrace",Null)
	Dim prefix As String = ""
	For i = 0 To StackTrace.Length - 1
		JO = StackTrace(i)
		Dim s As String = JO.RunMethod("toString",Null)
		If s.Contains("._" & startAtSub.toLowerCase) Then
			relevantSection = True
			Dim firstPeriod As Int = s.indexOf(".")
			prefix = s.SubString2(0, s.IndexOf2(".", firstPeriod + 1))
		End If
		If relevantSection And s.Contains("_") Then
			If s.Contains(prefix) = False Then Exit
			If s.contains("._")  = False Then s = s.SubString2(0, s.IndexOf("$")) & "." & s.SubString(s.IndexOf("_")).Replace(".resume", "")
			s = s.SubString(s.IndexOf("._") + 2)
			Dim subPart As String = s.SubString2(0, s.IndexOf("("))
			Dim contextPart As String = s.SubString2(s.IndexOf("(") + 1, s.IndexOf(".java:"))
			Dim p(2) As String
			p(0) = contextPart
			p(1) = subPart
			result.Add(p)
		End If
	Next
	If result.size > 0 Then result.RemoveAt(0)
	Dim sb As StringBuilder: sb.initialize
	For i = result.Size - 1 To 0 Step - 1
		Dim p() As String = result.Get(i)
		sb.Append(" → ").Append(p(0) & "." & p(1))
	Next
	Return sb.toString
End Sub

Private Sub toBBC(s As String) As String
	Dim v() As String = Regex.Split(CRLF, s)
	Dim sb As StringBuilder:sb.Initialize
	sb.Append("[Color=#0000ff]")
	Dim maxEntryLength As Int = -1
	For Each t As String In v
		Dim w() As String = Regex.Split(TAB, t)
		For Each g As String In w
			If g.Length > maxEntryLength Then maxEntryLength = g.length		
		Next
	Next
	Dim wx As Float = (12 * (maxEntryLength + 1))
	For Each t As String In v
		Dim w() As String = Regex.Split(TAB, t)
		
		For Each g As String In w
			sb.Append($"[Span MinWidth=${wx} Alignment=right]${g}[/Span]"$)
		Next
		sb.Append(CRLF)
	Next
	sb.Append("[/Color]")
	Return sb.toString
End Sub
