﻿B4J=true
Group=Default Group
ModulesStructureVersion=1
Type=Class
Version=8.5
@EndOfDesignText@
Sub Class_Globals
	Private Root As B4XView 'ignore
	#if B4J
		Private fx As JFX
	#End If
	Private xui As XUI 'ignore
	Private textEngine As BCTextEngine
	Private report As BBCodeView
	Private mData As dMatrix
	Private stats As WiLStats
	Private vars() As String
	Private MonoFont As B4XFont
	Private mCallBack As Object			'ignore
	
	Private MP As B4XMainPage
	Private dMx As dMatrix
End Sub

'Initializes object with variables of interest as columns of data
'the first column is the dependent variable, names is an Array of vraiable names
Public Sub Initialize (CallBack As Object, data As dMatrix, names() As String) As Object
	mData = data
	mCallBack = CallBack
	MP = B4XPages.MainPage

	stats.Initialize
	vars = names
	#if B4J
		Dim monospace As Font = fx.CreateFont("Consolas", 18, False, False)
	#Else If B4A
		Dim monospace As Typeface = Typeface.CreateNew(Typeface.MONOSPACE, Typeface.STYLE_NORMAL)
	#End If
	MonoFont = xui.CreateFont(monospace, 18)
	Return Me
End Sub

'This event will be called once, before the page becomes visible.
Private Sub B4XPage_Created (Root1 As B4XView)
	Root = Root1
	Root.LoadLayout("2")
	dMx.Initialize(MP, "")
	textEngine.Initialize(Root)
	textEngine.KerningEnabled = False
	textEngine.CustomFonts.Put("monospace", MonoFont)
	report.TextEngine = textEngine

#if B4J	
	Dim ts As Int = 15
#else if B4A
	Dim ts As Int = 16
#end if

	Dim nCases As Int = mData.nr
	Dim npreds As Int = mData.nc - 1
		'separate out the dependent variable, y=column(0) from the independent variables x=columns(1) to columns(npreds)
	Dim rowIndices() As Int = dMx.iInx(nCases, 0)
	Dim yColIndex() As Int = dMx.iInx(1, 0)
	Dim xColIndices() As Int = dMx.iInx(npreds, 1)
	Dim ones As dMatrix = dMx.dValue("C1", nCases, 1, 1)
	Dim y As dMatrix = mData.getSubset(rowIndices, yColIndex)
	Dim x As dMatrix = mData.getSubset(rowIndices, xColIndices).insertC(ones, 0)

	'the actual computation for the regression analysis is very concise - in 3 steps, "dot" is matrix multiplication
	Dim xt As dMatrix = x.transpose
	Dim K As dMatrix = xt.dot(x).inverse
	Dim B As dMatrix = K.dot(xt).dot(y)		'these are the coefficients of the prediction equation

	'compute the power of the predictive model - in essence correlate predicted y with observed y
	Dim predY As dMatrix = x.dot(B)
	Dim ypredy As dMatrix = y.appendC(predY)
	Dim R2 As Double = ypredy.correlationC.squared.x(0,1)
	Dim adjR2 As Double = 1 - (1 - R2) * (predY.nr - 1) / (predY.nr - x.nc)		'correct for having more than one predictor
	
	'now we need to compute the reliability of our prediction equation - this is presented as an ANOVA with an F-test
	Dim error As dMatrix = y.minus(predY)
	Dim errorMat As dMatrix = error.transpose.dot(error)
	errorMat.x(0, 0) = errorMat.x(0, 0) / (y.nr - npreds - 1)
	Dim ypredYerror As dMatrix = ypredy.appendC(error)
	Dim SSAnova As dMatrix = ypredYerror.varianceC(True).timesAmount(nCases - 1)
	Dim df As dMatrix = dMx.dNew("df", 1, 3, Array As Double(nCases - 1, 1 , nCases - npreds - 1))
	Dim MSS As dMatrix = SSAnova.divide(df)
	Dim FStatistic As Double = MSS.x(0,1) / MSS.x(0,2)
	Dim FProb As Double = 1 - stats.cdfF(FStatistic, df.x(0, 1), df.x(0, 2))

	'compute the standardized coefficients
	Dim seB As dMatrix = K.times(errorMat).getDiagonal.sqRoot
	Dim betas As dMatrix = B.times(x.stdDevC(True).divide(y.stdDevC(True)).transpose)
	
	'build the result matrix: Name, Estimate, Std. Error, beta, t-Value, p-Value
	Dim two0cols As dMatrix = dMx.dValue("C0", x.nc, 2, 0)
	Dim results As dMatrix = B.appendC(seB).appendC(betas).appendC(two0cols)

	'compute the reliability of the coefficients, add to result
	For i = 0 To B.nr - 1
		Dim t, p As Double
		p = 1
		If Abs(results.x(i, 1)) > .00000000001 Then
			t = results.x(i, 0) / results.x(i, 1)
			p = (1 - stats.cdfT(t, (y.nr - npreds - 2)))
		End If
		results.x(i, 3) = t
		results.x(i, 4) = p
	Next
	
	'prepare the report (BBCodeview) text
	Dim sb As StringBuilder: sb.initialize
	sb.Append("[b]" & vars(0) & "[/b]" & " = ")
	sb.Append(fmt(results.x(0, 0), 3))
	For i = 1 To results.nr - 1
		Dim val As String = fmt(Abs(results.x(i, 0)), 3)
		If results.x(i, 0) < 0 Then sb.Append(" - ") Else sb.Append(" + ")
		sb.Append(val).Append(" x ").Append("[b]" & vars(i) & "[/b]")
	Next
	Dim model As String = sb.ToString
	Dim prob As String = fmt(FProb, 5)
	If prob = "0.00000" Then prob = "<0.00001"
	
	'Code Snippet Alert: the following simplifies the TableRow generator
	Dim centerAlignments() As String = Array As String("center", "center", "center", "center", "center", "center")	'default empty array = center
	Dim plainEmphasis() As String = Array As String("plain")			'if only one item all columns receive the same value, empty array = plain
	Dim headerWidths() As Float = Array As Float(23, 12, 12, 12, 12, 12)
	Dim itemWidths() As Float = Array As Float(20, 12, 12, 12, 12, 12)
	Dim headers1() As String = Array As String("", "d.f.", "SS", "MSS", "F", "p-Value")
	Dim items1() As String = Array As String ("[b]Model[/b]", fmt(df.x(0,1), 0), fmt(SSAnova.x(0,1), 3), fmt(MSS.x(0,1), 3), fmt(FStatistic, 3), prob)
	Dim items2() As String = Array As String ("[b]Error[/b]", fmt(df.x(0,2), 0), fmt(SSAnova.x(0,2), 3), fmt(MSS.x(0,2), 3), "", "")
	Dim items3() As String = Array As String ("[b]Total[/b]", fmt(df.x(0,0), 0), fmt(SSAnova.x(0,0), 3), fmt(MSS.x(0,0), 3), "", "")
	Dim headers2() As String = Array As String("", "Estimate", "Std. Error", "beta", "t-Value", "p-Value")

	'this section depends on the number of predictors - so it has to be built line by line
	Dim sb As StringBuilder: sb.initialize
	For i = 0 To results.nr - 1
		Dim items(6) As String
		Dim prefix As String = "Amount"
		If i > 0 Then prefix = vars(i)
		items(0) = "[b]" & prefix & "[/b]"
		For j = 1 To 4
			items(j) = fmt(results.x(i, j - 1), 3)
		Next
		items(5) = fmt(results.x(i, 4), 5)
		If items(5) = "0.00000" Then items(5) = "<0.00001"						'zero probability is rare to find
		sb.Append(TableRow(items, itemWidths, Array As String("right"), plainEmphasis)).append(CRLF)
	Next

'putting it all together
	report.Text = $"
[font=monospace]
[Alignment=Center][TextSize=20][b]Multiple Linear Regression[/b][/TextSize][/Alignment]


[TextSize=${ts}][Alignment=Left]
     ${model}[/Alignment]
[/TextSize][TextSize=14]
      [b]Std. Error[/b] = ${fmt(Sqrt(errorMat.x(0, 0)), 3)}     [b]R-Sq[/b] = ${fmt(100*R2, 1)}%     [b]adjR-Sq[/b] = ${fmt(100*adjR2, 1)}%

${TableRow(headers1, headerWidths, centerAlignments, Array As String("bold"))}
${TableRow(items1, itemWidths, Array As String("right"), plainEmphasis)}
${TableRow(items2, itemWidths, Array As String("right"), plainEmphasis)}
${TableRow(items3, itemWidths, Array As String("right"), plainEmphasis)}

${TableRow(headers2, headerWidths, centerAlignments, Array As String("bold"))}
${sb.ToString}
"$
End Sub

Private Sub fmt(z As Double, nDecimal As Int) As String
	Return NumberFormat2(z, 1, nDecimal, nDecimal, False)
End Sub

'This routine facilitates the BBCodeview presentation of columns of items
Private Sub TableRow(columns() As String, colWidths() As Float, alignments() As String, emphasis() As String) As String
	Dim sb As StringBuilder: sb.Initialize
	For i = 0 To columns.Length - 1
		Dim c As String = columns(i)
		If c = "" Then c = Chr(160)
		Dim emp As String = "plain"
		If emphasis.length = 1 Then emp = emphasis(0)
		If emphasis.length = columns.Length Then emp = emphasis(i)
		If emp = "bold" Then c = $"[b]${c}[/b]"$
		Dim align As String = "center"
		If alignments.length = 1 Then align = alignments(0)
		If alignments.length = columns.Length Then align = alignments(i)
		Dim wx As Float = colWidths(i) * Root.Width / 100 / 1dip				'in B4A, dimensions are in dips
		sb.Append($"[Span MinWidth=${wx} Alignment=${align}]${c}[/Span]"$)
	Next
	Return sb.toString
End Sub

Private Sub Button1_Click
	B4XPages.ClosePage(Me)
End Sub
