﻿B4J=true
Group=Default Group
ModulesStructureVersion=1
Type=Class
Version=8.5
@EndOfDesignText@
Sub Class_Globals
	Private Root As B4XView 'ignore
	#if B4J
		Private fx As JFX
	#End If
	Private xui As XUI 'ignore
	Private textEngine As BCTextEngine
	Private report As BBCodeView
	Private mData As dMatrix
	Private mDictionary As Map
	Private stats As WiLStats
	Private MonoFont As B4XFont
	Private reportContent As StringBuilder
	Private maxEntryLength As Int
	Private opt As WiLOptions
	
	Private AN, ACorr As dMatrix
	Private vectors1, values1, components1, communalities1, rotatedFactors1, factorScores1, explanatoryPower As dMatrix
	Private nFactors As Int
	Private button1 As B4XView
	Private more As Label
	
	Private MP As B4XMainPage
	Private dMx As dMatrix
	Private dComp As decomposition
End Sub

'Initializes object with variables of interest as columns of data
' and dictionary that associates variable Vi with a description
Public Sub Initialize (CallBack As Object, data As dMatrix, dictionary As Map) As Object
	mData = data
	mDictionary = dictionary
	MP = B4XPages.MainPage
	stats.Initialize
	dComp.Initialize(Me, 3)
	#if B4J
		Dim monospace As Font = fx.CreateFont("Consolas", 18, False, False)
	#Else If B4A
		Dim monospace As Typeface = Typeface.CreateNew(Typeface.MONOSPACE, Typeface.STYLE_NORMAL)
	#End If
	MonoFont = xui.CreateFont(monospace, 18)
	Return Me
End Sub

'This event will be called once, before the page becomes visible.
Private Sub B4XPage_Created (Root1 As B4XView)
	Root = Root1
	Root.LoadLayout("2")
	Root.Color = xui.Color_RGB(220,220,220)
	button1.SetColorAndBorder(xui.Color_White, 1dip, xui.Color_Black, 5dip)
	
	dMx.Initialize(MP, "")

	textEngine.Initialize(Root)
	textEngine.KerningEnabled = False
	textEngine.CustomFonts.Put("monospace", MonoFont)
	report.TextEngine = textEngine
		
	AN  = mData.validPairsC
	ACorr = mData.correlationC
	
	Dim results() As dMatrix = dComp.eigenSymmetric(ACorr)
	vectors1 = results(0)
	values1 = results(1)
	
	components1 = components(vectors1, values1)
	communalities1 = communalities(components1)
	
	explanatoryPower = dMx.dValue("", 1, communalities1.nc, 0)
	nFactors = -1
	For j = 0 To communalities1.nc - 1
		Dim sum, num As Double
		For i = 0 To communalities1.nr - 1
			num = num + 1
			If communalities1.x(i, j) > .5 Then sum = sum + 1
		Next
		explanatoryPower.x(0, j) = 100 * sum / num
		If nFactors = -1 And explanatoryPower.x(0, j) > 80 Then nFactors = j + 1
	Next

	rotatedFactors1 = Varimax(components1)
	factorScores1 = factorScores(mData, rotatedFactors1)
End Sub

Private Sub reportPage(title As String, subTitle As String, rowNames As String, colNames As String, M As dMatrix, nInt As Int, nDecimal As Int, indices() As Int)
	If title <> "" Then title = $"${CRLF}[Alignment=Center][TextSize=20][b]${title}[/b][/TextSize][/Alignment]${CRLF}"$

#if B4J	
	Dim ts As Int = 15
#else if B4A
	Dim ts As Int = 16
#end if
	
	Dim headers(M.nc + 1) As String
	Dim cWidths(M.nc + 1) As Float
	Dim emph(M.nc + 1) As String 
	
	For j = 0 To M.nc
		headers(j) = colNames & j  
		emph(j) = "plain"
	Next
	headers(0) = ""
	emph(0) = "bold"
	Dim sb As StringBuilder: sb.Initialize
	Dim allEntries(M.nr, M.nc + 1) As String
	Dim saveMaxEntryLength As Int = maxEntryLength
	For i = 0 To M.nr - 1
		For j = 1 To M.nc
			allEntries(i, j) = NumberFormat2(M.x(i, j - 1), nInt, nDecimal, nDecimal, False)
			If subTitle.StartsWith("Explanatory") Then allEntries(i, j) = allEntries(i, j) & "%"
			Dim txt As Int = allEntries(i, j).Length
			If txt > maxEntryLength Then maxEntryLength = txt
		Next
	Next
	If maxEntryLength < saveMaxEntryLength Then maxEntryLength = saveMaxEntryLength
	
	Dim w As Float = 100 * 8.75 * maxEntryLength / Root.Width
	For j = 0 To M.nc
		cWidths(j) = w
	Next

	Dim entries(M.nc + 1) As String
	For i = 0 To M.nr - 1
		If rowNames <> "" Then entries(0) = rowNames & (1 + indices(i))
		For j = 1 To M.nc
			entries(j) = allEntries(i, j)	
			If subTitle.StartsWith("Rotated Factors") Then 
				If Abs(M.x(i, j - 1)) >= .45 Then 
					If M.x(i, j - 1) < 0 Then
						entries(j) = "[Color=#dd0000][b]" & entries(j) & "[/b][/Color]"
					Else
						entries(j) = "[Color=Blue][b]" & entries(j) & "[/b][/Color]"
					End If
				Else If Abs(M.x(i, j - 1)) >= .35 Then 
					If M.x(i, j - 1) < 0 Then
						entries(j) = "[Color=#dd0000][b]" & entries(j) & "[/b][/Color]"
					Else
						entries(j) = "[Color=#00aa00][b]" & entries(j) & "[/b][/Color]"
					End If
				Else If Abs(M.x(i, j - 1)) < .15 Then 
					entries(j) = "[Color=Gray]" & entries(j) & "[/Color]"
				End If
			Else If subTitle.StartsWith("Correlation") Then 
				If i = j - 1 Then entries(j) = "[Color=Gray]1.0[/Color]"
			Else If subTitle.StartsWith("Valid") Then 
				If i = j - 1 Then entries(j) = "[Color=Gray]" & entries(j) & "[/Color]"
			End If
		Next
		Dim rowContent As String = TableRow(entries, cWidths, Array As String("right"), emph)
		If subTitle.StartsWith("Rotated Factors") Then 
			rowContent.SubString2(0, rowContent.Length -1)
			sb.Append(CRLF).Append($"[TextSize=${ts}]${rowContent}[/TextSize][TextSize=12] ${mDictionary.Get(entries(0))}[/TextSize]"$)
		Else
			sb.Append(CRLF).Append($"[TextSize=${ts}]${rowContent}[/TextSize]"$)
		End If
	Next
	
	Dim note As String = ""
	If subTitle.Contains("(") Then 
		Dim v() As String = twoParts(subTitle, "(")
		subTitle = v(0)
		note =  $"[TextSize=${ts}][Color=Black](${v(1)}[/Color][/TextSize]"$
	End If

	reportContent.Append($"
[font=monospace]${title}
[Indent=3][TextSize=18][b][Color=#0000ff]${subTitle}[/Color][/b][/TextSize]${note}
[Indent=0][TextSize=${ts}]${TableRow(headers, cWidths, Array As String("center"), Array As String("bold"))}[/TextSize]${sb.ToString}
"$)
	report.Text = reportContent.toString
	Dim ph As Float = report.Paragraph.Height - 40dip
	If ph / textEngine.mScale + report.Padding.Top + report.Padding.Bottom > report.mBase.height Then 
		more.visible = True
	Else
		more.visible = False
	End If
End Sub

Private Sub twoParts(s As String, c As String) As String()
	Dim r(2) As String
	r(0) = s
	r(1) = ""
	Dim k As Int = s.IndexOf(c)
	If k > -1 Then
		r(0) = s.SubString2(0, k)
		r(1) = s.SubString(k + c.Length)
	End If
	Return r
End Sub

'This routine facilitates the BBCodeview presentation of columns of items
Private Sub TableRow(columns() As String, colWidths() As Float, alignments() As String, emphasis() As String) As String
	Dim sb As StringBuilder: sb.Initialize
	For i = 0 To columns.Length - 1
		Dim c As String = columns(i)
		If c = "" Then c = Chr(160)
		Dim emp As String = "plain"
		If emphasis.length = 1 Then emp = emphasis(0)
		If emphasis.length = columns.Length Then emp = emphasis(i)
		If emp = "bold" Then c = $"[b]${c}[/b]"$
		Dim align As String = "center"
		If alignments.length = 1 Then align = alignments(0)
		If alignments.length = columns.Length Then align = alignments(i)
		Dim wx As Float = colWidths(i) * Root.Width / 100 
		sb.Append($"[Span MinWidth=${wx} Alignment=${align}]${c}[/Span]"$)
	Next
	Return sb.toString
End Sub

Private Sub b4xPage_Appear
	Button1_Click
End Sub

Private Sub Button1_Click
	reportContent.initialize
	report.Text = $"
[font=monospace]
[Alignment=Center][TextSize=20][b]PCA and Factor Analysis[/b][/TextSize][/Alignment]

"$
	button1.visible = False
	Dim options() As String = Array As String("Correlation Matrix", "Eigen Vectors and Values", "Principal Components", _
	"Variance Explained", "Rotated Factors", "Weighted Factor Scores")
	opt.Initialize(Root, "Select a Report")
	Dim dia As Object = opt.show(options)
	Wait for (dia) complete (result As Int)
	Dim choice As String
	If result = xui.DialogResponse_Positive Then choice = opt.listTemplate.SelectedItem
	button1.visible = True
	maxEntryLength = 0
	Dim indices() As Int = dMx.iInx(ACorr.nc, 0)
	Dim title As String = "Principal Component and Factor Analysis"
	Dim subtitle As String
	Select choice
		Case "Correlation Matrix"
			subtitle = "Correlation Matrix"
			reportPage(title, subtitle, "v","v", ACorr, 0, 3, indices)
			subtitle = "Valid Number of Cases"
			reportPage("", subtitle, "v", "v", AN, 1, 0, indices)
		Case "Eigen Vectors and Values"
			subtitle = "Eigen Vectors"
			reportPage(title, subtitle, "v", "e", vectors1, 0, 3, indices)
			subtitle = "Eigen Values"
			reportPage("", subtitle, "", "v", values1.getDiagonal.transpose, 0, 3, indices)
		Case "Principal Components"
			subtitle = "Principal Components"
			reportPage(title, subtitle, "v", "c", components1, 0, 3, indices)
		Case "Variance Explained"
			subtitle = "Communalities (cumulative proportion of variance explained by factor)"
			reportPage(title, subtitle, "v", "f", communalities1, 0, 3, indices)
			subtitle = "Explanatory Power (% of variables with > 50% of variance explained)"
			reportPage("", subtitle, "", "f", explanatoryPower, 0, 0, indices)
			subtitle = "Number of Factors = " & nFactors & " (based on explanatory power > 80%)"
			Dim recommend As dMatrix = explanatoryPower.copy
			recommend.nr = 0
			recommend.nc = 0
			reportPage("", subtitle, "", "", recommend, 0, 0, indices)
		Case "Rotated Factors"			
			subtitle = "Rotated Factors (Target Number = " & nFactors & ")"
			Dim augRF As dMatrix = dMx.dValue("", rotatedFactors1.nr, rotatedFactors1.nc + 2, 0)
			For i = 0 To rotatedFactors1.nr - 1
				augRF.x(i, 0) = i
				Dim sum As Double
				For j = 0 To rotatedFactors1.nc - 1
					Dim k As Int = rotatedFactors1.nc - j
					Dim z As Double = rotatedFactors1.x(i, j)
					Dim cat As Float = 0
					If z > .45 Then cat = Power(10, k) + z
					sum = sum + cat
					augRF.x(i, j+2) = z
				Next
				augRF.x(i, 1) = sum
			Next
			augRF = augRF.sortC(1, False)
			Dim augRFSubset As dMatrix = dMx.dValue("", augRF.nr, augRF.nc - 2, 0)
			For i = 0 To rotatedFactors1.nr - 1
				indices(i) = augRF.x(i, 0)
				For j = 0 To augRF.nc - 3
					augRFSubset.x(i, j) = augRF.x(i, j + 2)
				Next
			Next
			reportPage(title, subtitle, "v", "f", augRFSubset, 0, 3, indices)
		Case "Weighted Factor Scores"
			subtitle = "Weighted Factor Scores (Weights: 1 if loading > .5, else 0)"
			reportPage(title, subtitle, "", "f", factorScores1, 0, 3, indices)
		Case ""
			B4XPages.ClosePage(Me)
	End Select
End Sub

'computes and returns communalities as a dMatrix object
Public Sub communalities(mat As dMatrix) As dMatrix
	Dim communal As dMatrix = mat.copy.rename("communalities")
	For i = 0 To communal.nr - 1
		Dim sum As Double = 0
		For j = 0 To communal.nc - 1
			sum = sum + communal.x(i, j) * communal.x(i, j)
			communal.x(i, j) = sum
		Next
	Next
	Return communal
End Sub

'computes and returns principal components as a dMatrix object
Public Sub components(mat As dMatrix, values As dMatrix) As dMatrix
	Dim eps As Double = Power(2.0, -52.0)
	For i = 0 To values.nr - 1
		If values.x(i, i) < eps Then 
			Log("Matrix is unstable - Eigenvalues include a negative or very small number")
			Return mat
		End If
	Next
	Dim valuesSqrt As dMatrix = values.sqRoot
	Dim comps As dMatrix = mat.dot(valuesSqrt).rename("components")
	Return comps
End Sub

'computes and returns rotated factors as a dMatrix object
Public Sub Varimax(mat As dMatrix) As dMatrix
	'Translated to B4X from MATLAB procedure which was modified by Trevor Park in April 2002
	'	 from an original file by J.O. Ramsay - a prolific author in Multivariate Analysis

	Private cMx As cMatrix: cMx.Initialize(Me, "")
	Dim nIters As Int = 30
	Dim n As Int = mat.nr
	Dim m As Int = nFactors
	Dim rowIndices() As Int = dMx.iInx(n, 0)
	Dim lambda As dMatrix = mat.getSubset(rowIndices, dMx.iInx(m, 0))
	If nFactors = 1 Then Return lambda
	
	Dim H As dMatrix = mat.squared.sumR.sqRoot
	Dim HInv As dMatrix = dMx.dValue("", n, 1, 1).divide(H)
	Dim LambdaStar As dMatrix = lambda.times(HInv).rename("LambdaStar")
	Dim lastSV, P, FJ, FK, cosP, sinP As Double
	For iter = 0 To nIters - 1
		For j = 0 To m - 2
			For k = j + 1 To m - 1
				'Calculate optimal 2-D planar rotation angle for columns j,k
				Dim colPair, colPair2, term1, term2 As cMatrix
				colPair = cMx.cPairs(LambdaStar, j, k)
				colPair2 = colPair.squared
				term1 = colPair2.sumC.squared
				term2 = colPair2.squared.sumC.timesAmount(n)
				Dim difSums As cMatrix = term2.minus(term1)
				Dim P As Double = difSums.phase.dMx_R.timesAmount(.25).x(0,0)
				sinP = Sin(P)
				If Abs(sinP) > Power(10, -5 * dComp.nDecimals) Then 
					cosP = Cos(P)
					For i = 0 To n - 1
						FJ = LambdaStar.x(i, j): FK = LambdaStar.x(i, k)
						LambdaStar.x(i, j) = FJ * cosP + FK * sinP
						LambdaStar.x(i, k) = -FJ * sinP + FK * cosP
					Next
				End If
			Next
		Next
		Dim SV As Double = LambdaStar.squared.varianceC(False).sumR.x(0, 0)
		If Abs(SV - lastSV) <= Power(10, -5 * dComp.nDecimals) Then Exit
		lastSV = SV 
	Next
	lambda = LambdaStar.times(H)
	lambda = lambda.iterate(lambda, "reflect", Array(lambda.sumC, "row")).rename("rotatedFactors")
	If iter = nIters Then Log("No convergence")
	Return lambda
End Sub

'computes and returns rotated factors scores as a dMatrix object
'the algorithm corrects for different scaled variables and number of meaning loadings a factor has
'the scores are scale-free and indicate "relative" link to a given rotated factor
Public Sub factorScores(mat As dMatrix, rFactors1 As dMatrix) As dMatrix
	Dim nf As Int = rFactors1.nc
	Dim nv As Int = rFactors1.nr
	Dim ncases As Int = mat.nr
	Dim fScores As dMatrix = dMx.dValue(mat.id & ".factorScores", ncases, nf, 0)
	Dim wghts As dMatrix = rFactors1.iterate(rFactors1, "compareGT", Array(.5))
	wghts = wghts.divide(wghts.SumC)		'how many of possible positives for a factor are positive?
	Dim mins As dMatrix = mat.minC
	Dim maxs As dMatrix = mat.maxC
	For i = 0 To ncases - 1
		For k = 0 To nf - 1
			Dim sum, num As Double
			For j = 0 To nv - 1
				If mat.x(i, j) = mat.x(i, j) Then 
					sum = sum + wghts.x(j, k) * (mat.x(i, j) - mins.x(0, j))/ (maxs.x(0, j) - mins.x(0, j))
					num = num + 1
				End If
			Next
			Dim factorSc As Double
			If num > 0 Then factorSc = sum / num
			fScores.x(i, k) = 100 * factorSc
		Next	
	Next
	Return fScores
End Sub
