﻿B4J=true
Group=Default Group
ModulesStructureVersion=1
Type=Class
Version=8.5
@EndOfDesignText@
'Introduction:

'This class does only one thing - replace the lookup tables used in statistics with continuous functions
'To determine if a Correlation, ChiSq, F, t, or z-value are significant, the exact(almost) cumulative probability is computed
'The calling program needs to subtract that from 1.000 to arrive at the classical p-value

'It is not necessary to understand the algorithms in order to use them
'Check Wiki for definitions of Cumulative (Probability) Density Functions, Gamma, Beta, and Continuous Fractions

Sub Class_Globals
#if B4J
	Private fx As JFX
#End If
End Sub

Public Sub Initialize
End Sub

'returns the hyperbolic tangent  
Public Sub tanh(xs As Float) As Float
	Dim term As Double = Power(cE, -2*xs)
	Return (1-term)/(1+term)
End Sub

'returns the hyperbolic arctangent  
Public Sub atanh(xs As Float) As Float
	If Abs(xs)<1 Then
		Dim term As Double = (1 + xs)/(1 - xs)
		Return .5*Logarithm(term, cE)
	Else
		Return 0
	End If
End Sub

'returns the cumulative proportion of the area under then Normal distribution (two-tailed)
Public Sub cdfZ(xs As Double) As Double
	If xs >= 0 Then Return (gammaNormal(.5, xs * xs / 2)) Else Return cdfZ(-xs)
End Sub

'returns the cumulative proportion of the area under then Chi-squared distribution (always one-tailed)
Public Sub cdfC(xs As Double, v As Double) As Double
	If v < 1 Then Return 0
	If xs > 0 Then Return (gammaNormal(.5 * v, xs / 2)) Else Return 0
End Sub

'returns the cumulative proportion of the area under then Student t-distribution (two-tailed)
Public Sub cdfT(xs As Double, v As Double) As Double
	If v < 1 Then Return 0
	If xs>=0 Then Return (1 - betaNormal(v / 2, .5, v / (v + xs*xs))) Else Return cdfT(-xs, v)
End Sub

'returns the cumulative proportion of the area under then F-distribution (always one-tailed)
Public Sub cdfF(xs As Double, v1 As Double, v2 As Double) As Double
	If v1 < 1 Or v2 < 1 Then Return 0
	If xs>0 Then Return (1 - betaNormal(v2 / 2, v1 / 2, v2 / (v2 + v1*xs))) Else Return 0
End Sub

'returns the cumulative proportion of the area under then Student t-distribution (two-tailed)
' after the Pearson correlation coefficient is transformed to t
Public Sub cdfR(xs As Double, v As Double) As Double
	If v < 1 Then Return 0
	If xs>=0 Then Return cdfT(xs * Sqrt(v)/Sqrt(1.0 - xs*xs), v) Else Return cdfR(-xs, v)
End Sub

Private Sub gammaNormal(a As Double, xs As Double) As Double
	Dim sum, delta, aStep, result As Double
	result = 0.0
	If (xs = 0.0) Then
		result = .5
	Else
		aStep = a: delta = 1.0 / a
		sum = delta
		Dim n As Int = 0
		Do While (Abs(delta) > Abs(sum) * .0000000001)
			n = n + 1
			If n > 250 Then Exit
			aStep = aStep + 1
			delta = delta * xs / aStep
			sum = sum + delta
		Loop
		result = sum * Power(cE, -xs + a * Logarithm(xs,cE) - logGamma(a))
	End If
	Return result
End Sub

Private Sub betaNormal(a As Double, b As Double, xs As Double) As Double
	If (xs < 0.0 Or xs > 1.0) Then Return 0.0
	Dim result As Double
	If (xs = 0.0 Or xs = 1.0) Then
		result = 0.0
	Else
		result = Power(cE, logGamma(a + b) - logGamma(a) - logGamma(b) + a * Logarithm(xs, cE) + b * Logarithm(1.0 - xs, cE))
	End If
	If (xs < (a + 1.0)/(a + b + 2.0)) Then
		result = result * continuousFractions(a, b, xs) / a
	Else
		result =  1.0 - result * continuousFractions(b, a, 1.0-xs) / b
	End If
	Return result
End Sub

Private Sub continuousFractions(a As Double, b As Double, xs As Double) As Double
	Dim m, m2 As Int
	Dim aa, c, d, delta, result As Double
	c = 1.0
	d = 1.0 / Abs(1.0 - (a +  b) * xs / (a + 1))
	result = d
	m = 0: m2 = 0
	Do While Abs(delta-1.0) > .00000000001
		m = m + 1
		If m>100 Then Return 0.0
		m2 = m2 + 2
		aa = m * (b - m) * xs / ((a + m2) * (a - 1 + m2))
		c = (1.0 + aa/c)
		d = 1.0 / (1.0 + aa*d)
		result = result * c * d
		aa = -(a + m) * (a + b + m) * xs / ((a + m2) * (a + 1 + m2))
		c = (1.0 + aa/c)
		d = 1.0 / (1.0 + aa*d)
		delta = c * d
		result = result * delta
	Loop
	Return result
End Sub

Private Sub logGamma(z As Double) As Double
	Dim  zStep, denominator, series, Lanczos() As Double   'Lanzos coefficients g=5 n=7
	Lanczos = Array As Double( 1.000000000190015, 76.18009172947146, -86.50532032941677, 24.01409824083091, -1.231739572450155, 0.1208650973866179e-2, -0.5395239384953e-5)
	denominator = (z + 5.5) - (z + 0.5) * Logarithm((z + 5.5), cE)
	series = Lanczos(0)
	zStep = z
	For j = 1 To 6
		zStep = zStep + 1
		series = series + Lanczos(j) / zStep
	Next
	Return Logarithm(Sqrt(2*cPI) * series / z, cE) - denominator
End Sub
