﻿B4J=true
Group=Default Group
ModulesStructureVersion=1
Type=Class
Version=8.5
@EndOfDesignText@
'NOTE: This class is dependent on the dMatrix class.

'Complex numbers are implemented here as paired, synchronized double precision matrices, with same shape
'It uses the methods of dMatrix extensively

Sub Class_Globals
	Public id As String
	Public nr, nc As Int
	Public dMx_R As dMatrix
	Public dMx_I As dMatrix
	Public firstMethodApplied As String
	Public lastMethodApplied As String
	
	Private mCallBack As Object
	Private dMx As dMatrix
End Sub

'Global: Private cMx as cMatrix  cMx.Initialize(Me, "")  provides accesses to the various contructors and utilities
Public Sub Initialize(callBack As Object, name As String) As cMatrix
	dMx.Initialize(callBack, "")
	mCallBack = callBack
	id = name
	nr = 0: nc = 0
	dMx_R.Initialize(mCallBack, "")
	dMx_I.Initialize(mCallBack, "")
	Return Me
End Sub

Private Sub methods(M As cMatrix, method As String) As cMatrix
	If firstMethodApplied = "" Then 
		M.firstMethodApplied = "." & method
		M.lastMethodApplied = ""
	Else 
		M.firstMethodApplied = firstMethodApplied
		M.lastMethodApplied = "…" & method
	End If
	For i = 0 To M.nr - 1
		For j = 0 To M.nc - 1 	
			If (M.dMx_R.x(i,j) <> M.dMx_R.x(i,j)) Or (M.dMx_I.x(i,j) <> M.dMx_I.x(i,j)) Then 
				M.dMx_R.x(i,j) = "NaN"
				M.dMx_I.x(i,j) = "NaN"
			End If
		Next
	Next
	Return M
End Sub

'"id" property of dMatrix is used for display and tracing errors, rename resets old traces
Public Sub rename(name As String) As cMatrix
	id = name
	firstMethodApplied = ""
	lastMethodApplied = ""
	Return Me
End Sub

'cNew method is the most generic constructor of dMatrix, came becomes "id"
Public Sub cNew(name As String, m As Int, n As Int, value_R() As Double, value_I() As Double) As cMatrix
	Dim newM As cMatrix: newM.Initialize(mCallBack, name)
	newM.dMx_R = dMx.dNew("", m, n, value_R)
	newM.dMx_I = dMx.dNew("", m, n, value_I)
	newM.nr = newM.dMx_R.nr: newM.nc = newM.dMx_R.nc
	Return newM
End Sub

'cValue method is a common constructor of cMatrix, usually to prepare a Amount (real and imaginary) filled matrix of the right shape
Public Sub cValue(name As String, m As Int, n As Int, value_R As Double, value_I As Double) As cMatrix
	Return cNew(name, m, n, Array As Double(value_R), Array As Double(value_I))
End Sub

'cZero method is a common constructor of cMatrix, usually to prepare a zero filled matrix of the right shape
Public Sub cZero(name As String, m As Int, n As Int) As cMatrix
	Return cNew(name, m, n, Array As Double(0), Array As Double(0))
End Sub

'cScalar constructs a 1x1 matrix with "z" as the entry
Public Sub dScalar(name As String, value_R As Double, value_I As Double) As cMatrix
	Return cNew(name, 1, 1, Array As Double(value_R), Array As Double(value_I))
End Sub

'cDiagonal constructs a diagonal matrix with the items in the double array v() cyclically filling the diagonal
Public Sub cDiagonal(name As String, n As Int, value_R As Double, value_I As Double) As cMatrix
	Dim newM As cMatrix: newM.Initialize(mCallBack, name)
	newM.nr = n: newM.nc = dMx_R.nc
	nr = n: nc = n
	dMx_R = dMx.dDiagonal("", n, Array As Double(value_R))
	dMx_I = dMx.dDiagonal("", n, Array As Double(value_I))
	Return Me
End Sub

'cIdentity constructs a diagonal matrix with the ones on the diagonal
Public Sub cIdentity(name As String, m As Int) As cMatrix
	Return cDiagonal(name, m, 1, 0)
End Sub

'cString takes a smart string and parses it into a matrix that is wrapped in a cMatrix object
Public Sub cString(name As String, string_R As String, string_I As String) As cMatrix
	Dim newM As cMatrix: newM.Initialize(mCallBack, name)
	newM.dMx_R = dMx.dString("", string_R)
	newM. dMx_I = dMx.dString("", string_I)
	newM.nr = newM.dMx_R.nr: newM.nc = newM.dMx_R.nc
	Return newM
End Sub

'cRandom constructs a cMatrix object with entries that are random numbers between lower and upper with nDecimal places
Public Sub cRandom(name As String, m As Int, n As Int, lowerR As Double, upperR As Double, lowerI As Double, upperI As Double, nDecimal As Int) As cMatrix
	Dim newM As cMatrix: newM.Initialize(mCallBack, name)
	newM.dMx_R = dMx.dRandom("", m, n, lowerR, upperR, nDecimal)
	newM.dMx_I = dMx.dRandom("", m, n, lowerI, upperI, nDecimal)
	newM.nr = newM.dMx_R.nr: newM.nc = newM.dMx_R.nc
	Return newM
End Sub

'cWrap takes an ordinary matrix and constructs a cMatrix object
Public Sub cWrap(name As String, matR(,) As Double, matI(,) As Double) As cMatrix
	Dim newM As cMatrix: newM.Initialize(mCallBack, name)
	newM.dMx_R = dMx.dWrap("", matR)
	newM.dMx_I = dMx.dWrap("", matI)
	newM.nr = newM.dMx_R.nr: newM.nc = newM.dMx_R.nc
	Return newM
End Sub

'cList takes a two equally sized lists of vectors (all same length) and constructs a cMatrix object
Public Sub dList(name As String, aList As List, bList As List) As cMatrix
	Dim newM As cMatrix: newM.Initialize(mCallBack, name)
	newM.dMx_R = dMx.dList("", aList)
	newM.dMx_I = dMx.dList("", bList)
	newM.nr = newM.dMx_R.nr: newM.nc = newM.dMx_R.nc
	Return newM
End Sub

'cPairs constructs a cMatrix object out of two columns of "mat"
Public Sub cPairs(mat As dMatrix, i As Int, j As Int) As cMatrix
	Dim newM As cMatrix: newM.Initialize(mCallBack, id & "pairs")
	newM.dMx_R = mat.getColumn(i)
	newM.dMx_I = mat.getColumn(j)
	newM.nr = newM.dMx_R.nr: newM.nc = newM.dMx_R.nc
	Return newM
End Sub

'For flexibility (Pages or no Pages), the output is handled by "myLog" which has to be in the calling program
Public Sub display(nDecimal As Int) As cMatrix
	Dim eps As Double = 1 / Power(10, nDecimal + 2)		'used to avoid the minus sign in -0.0000
	Dim sb As StringBuilder: sb.initialize
	For i = 0 To nr - 1
		For j = 0 To nc - 1
			Dim valR As Double = dMx_R.x(i, j)
			If Abs(valR) < eps Then valR = 0
			Dim sR As String = NumberFormat2(valR, 1, nDecimal, nDecimal, False)
			If sR = "NaN" Then sR = Chr(0x25C8)
			Dim valI As Double = dMx_I.x(i, j)
			If Abs(valI) < eps Then valI = 0
			Dim sI As String = NumberFormat2(valI, 1, nDecimal, nDecimal, False)
			If sI = "NaN" Then sI = Chr(0x25C8)
			Dim complex As String = $"${sR}+${sI}i"$.replace("+-", "-")
			sb.Append(complex).Append(TAB)
		Next
		sb.Remove(sb.Length -1, sb.Length).append(CRLF)
	Next
	CallSub3(mCallBack, "myLog", sb.toString, $"${id}${firstMethodApplied}${lastMethodApplied} [${nr}, ${nc}]"$)
	Return Me
End Sub

'copies all aspects of the cMatrix object to a new cMatrix object
Public Sub copy As cMatrix
	Dim newM As cMatrix: newM.Initialize(mCallBack, id)
	newM.dMx_R = dMx_R.copy
	newM.dMx_I = dMx_I.copy
	newM.nr = newM.dMx_R.nr: newM.nc = newM.dMx_R.nc
	Return methods(newM, "copy")
End Sub

'getDiagonal selects the diagonal and constructs a min(m,n) x 1 cMatrix object
Public Sub getDiagonal As cMatrix
	Dim newM As cMatrix: newM.Initialize(mCallBack, id)
	newM.dMx_R = dMx_R.getDiagonal
	newM.dMx_I = dMx_I.getDiagonal
	newM.nr = newM.dMx_R.nr: newM.nc = newM.dMx_R.nc
	Return methods(newM, "diagonal")
End Sub

'getColumn selects the intended column and constructs a 1xn cMatrix object
Public Sub getColumn(index As Int) As cMatrix
	Dim newM As cMatrix: newM.Initialize(mCallBack, id)
	newM.dMx_R = dMx_R.getColumn(index)
	newM.dMx_I = dMx_I.getColumn(index)
	newM.nr = newM.dMx_R.nr: newM.nc = newM.dMx_R.nc
	Return methods(newM, "column" & index)
End Sub

'getRow selects the intended row and constructs a mx1 cMatrix object
Public Sub getRow(index As Int) As cMatrix
	Dim newM As cMatrix: newM.Initialize(mCallBack, id)
	newM.dMx_R = dMx_R.getRow(index)
	newM.dMx_I = dMx_I.getRow(index)
	newM.nr = newM.dMx_R.nr: newM.nc = newM.dMx_R.nc
	Return methods(newM, "row" & index)
End Sub

'computes the complex matrix "abs" contruct and returns a new cMatrix object
Public Sub absValue As cMatrix
	Dim newM As cMatrix = copy
	newM.dMx_R = dMx_R.squared.plus(dMx_I.squared).sqRoot
	newM.dMx_I = dMx.dValue(id, nr, nc, 0)
	Return methods(newM, "absValue")
End Sub

'computes the complex matrix "phase" contruct and returns a new cMatrix object
Public Sub phase As cMatrix
	Dim newM As cMatrix = copy
	For i = 0 To nr - 1
		For j = 0 To nc - 1
			Dim x As Double = newM.dMx_R.x(i, j)
			Dim y As Double = newM.dMx_I.x(i, j)
			If x>0 Or y<> 0 Then 
				newM.dMx_R.x(i, j) = 2 * ATan(y/(Sqrt(x*x +y*y) + x))
			Else if x<0 And y = 0 Then
				newM.dMx_R.x(i, j) = cPI
			Else if x = 0 And y = 0 Then 
				newM.dMx_R.x(i, j) = "NaN"
			End If
		Next
	Next
	newM.dMx_I = dMx.dValue(id, nr, nc, 0)
	Return methods(newM, "phase")
End Sub

'computes the complex matrix "conjugate" contruct (imaginary part is negated) and returns a new cMatrix object
Public Sub conjugate As cMatrix
	Dim newM As cMatrix = copy
	newM.dMx_I = dMx_I.negate
	Return methods(newM, "conjugate")
End Sub

'adds the elements of a cMatrix object (M) and returns a new cMatrix object
'M can must have commensurate dimensions (possibly single row, single column, or 1x1)
Public Sub plus(mat As cMatrix) As cMatrix
	Dim newM As cMatrix = copy
	newM.dMx_R = dMx_R.plus(mat.dMx_R)
	newM.dMx_I = dMx_I.plus(mat.dMx_I)
	Return methods(newM, "plus." & mat.id)
End Sub

'muliplies elements by a elements of cMatrix object (M) and returns a new cMatrix object
'M can must have commensurate dimensions (possibly single row, single column, or 1x1)
Public Sub times(mat As cMatrix) As cMatrix
	Dim newM As cMatrix = copy
	newM.dMx_R = dMx_R.times(mat.dMx_R).minus(dMx_I.times(mat.dMx_I))
	newM.dMx_I = dMx_R.times(mat.dMx_I).plus(dMx_I.times(mat.dMx_R))
	Return methods(newM, "times." & mat.id)
End Sub

'subtracts elements by elements of a cMatrix object (M) and returns a new cMatrix object
'M can must have commensurate dimensions (possibly single row, single column, or 1x1)
Public Sub minus(mat As cMatrix) As cMatrix
	Dim newM As cMatrix = copy
	newM.dMx_R = dMx_R.minus(mat.dMx_R)
	newM.dMx_I = dMx_I.minus(mat.dMx_I)
	Return methods(newM, "minus." & mat.id)
End Sub

'squares the complex elements and returns a cMatrix object with the results
Public Sub squared As cMatrix
	Dim newM As cMatrix = copy
	newM.dMx_R = dMx_R.times(dMx_R).minus(dMx_I.times(dMx_I))
	newM.dMx_I = dMx_R.times(dMx_I).plus(dMx_I.times(dMx_R))
	Return methods(newM, "squared")
End Sub

'computes the complex reciprocal of each element and returns a cMatrix object with the results
Public Sub reciprocal As cMatrix
	Dim newM As cMatrix = copy
	newM.dMx_R = dMx_R.divide(dMx_R.squared.plus(dMx_I.squared))
	newM.dMx_I = dMx_I.divide(dMx_R.squared.plus(dMx_I.squared)).negate
	Return methods(newM, "reciprocal")
End Sub

'divides each element (multiplies by complex reciprocal) by corresponding element of "mat" and returns a cMatrix object with the results
Public Sub divide(mat As cMatrix) As cMatrix
	Dim newM As cMatrix = times(mat.reciprocal)
	Return methods(newM, "divide." & mat.id)
End Sub

'adds a amount to each element and returns a cMatrix object with the results
Public Sub plusAmount(z As Double) As cMatrix
	Dim newM As cMatrix = copy
	newM.dMx_R = dMx_R.plusAmount(z)
	newM.dMx_I = dMx_I.plusAmount(z)
	Return methods(newM, "plus" & z)
End Sub

'multiplies each by an amount and returns a cMatrix object with the results
Public Sub timesAmount(z As Double) As cMatrix
	Dim newM As cMatrix = copy
	newM.dMx_R = dMx_R.timesAmount(z)
	newM.dMx_I = dMx_I.timesAmount(z)
	Return methods(newM, "times" & z)
End Sub

'Transforms construct a cMatrix object with row and columns transposed newX(i, j) = oldX(j, i)
Public Sub transpose As cMatrix
	Dim newM As cMatrix: newM.Initialize(mCallBack, id)
	newM.dMx_R = dMx_R.transpose
	newM.dMx_I = dMx_I.transpose
	newM.nr = newM.dMx_R.nr: newM.nc = newM.dMx_R.nc
	Return methods(newM, "transpose")
End Sub

'computes the sum of each column (adjusting for missing values) and returns a 1xn cMatrix object
Public Sub sumC As cMatrix
	Dim newM As cMatrix: newM.Initialize(mCallBack, id)
	newM.dMx_R = dMx_R.sumC
	newM.dMx_I = dMx_I.sumC
	newM.nr = newM.dMx_R.nr: newM.nc = newM.dMx_R.nc
	Return methods(newM, "sumC")
End Sub


'computes the sum of each row (adjusting for missing values) and returns a mx1 cMatrix object
Public Sub sumR As cMatrix
	Dim newM As cMatrix: newM.Initialize(mCallBack, id)
	newM.dMx_R = dMx_R.sumR
	newM.dMx_I = dMx_I.sumR
	newM.nr = newM.dMx_R.nr: newM.nc = newM.dMx_R.nc
	Return methods(newM, "sumR")
End Sub

'inserts a single column cMatrix object into column index position - returns a new cMatrix object
Public Sub insertC(mat As cMatrix, position As Int) As cMatrix
	Dim newM As cMatrix: newM.Initialize(mCallBack, id)
	newM.dMx_R = dMx_R.insertC(mat.dMx_R, position)
	newM.dMx_I = dMx_I.insertC(mat.dMx_I, position)
	newM.nr = newM.dMx_R.nr: newM.nc = newM.dMx_R.nc
	Return methods(newM, "insertC" & position & "." & mat.id)
End Sub


'inserts a single row cMatrix object into row index position - returns a new cMatrix object
Public Sub insertR(mat As cMatrix, position As Int) As cMatrix
	Dim newM As cMatrix: newM.Initialize(mCallBack, id)
	newM.dMx_R = dMx_R.insertR(mat.dMx_R, position)
	newM.dMx_I = dMx_I.insertR(mat.dMx_I, position)
	newM.nr = newM.dMx_R.nr: newM.nc = newM.dMx_R.nc
	Return methods(newM, "insertR" & position & "." & mat.id)
End Sub


'Appends a cMatrix object with the same number of rows - returns a new cMatrix object
Public Sub appendC(mat As cMatrix) As cMatrix
	Dim newM As cMatrix: newM.Initialize(mCallBack, id)
	newM.dMx_R = dMx_R.appendC(mat.dMx_R)
	newM.dMx_I = dMx_I.appendC(mat.dMx_I)
	newM.nr = newM.dMx_R.nr: newM.nc = newM.dMx_R.nc
	Return methods(newM, "appendC" & "." & mat.id)
End Sub


'Appends a cMatrix object with the same number of columns - returns a new cMatrix object
Public Sub appendR(mat As cMatrix) As cMatrix
	Dim newM As cMatrix: newM.Initialize(mCallBack, id)
	newM.dMx_R = dMx_R.appendR(mat.dMx_R)
	newM.dMx_I = dMx_I.appendR(mat.dMx_I)
	newM.nr = newM.dMx_R.nr: newM.nc = newM.dMx_R.nc
	Return methods(newM, "appendR" & "." & mat.id)
End Sub

'Matrix multiplication "dot" is the matrix operator (rows of the 1st matrix are multiplied by columns of the 2nd matrix and summed)
'It is named after "dot product" in vector notation - the row vectors of matrix1 are "dot product"ed with the columns of matrix2
'constraints - the #columns of the first has to be equal to the #rows of the second matrix
Public Sub dot(M As cMatrix) As cMatrix
	Dim newM As cMatrix = cValue(id, nr, M.nc, 0, 0)
	Dim sum(2), num As Double
	For i = 0 To newM.nr - 1
		For j = 0 To newM.nc - 1
			sum(0) = 0: sum(1) = 0: num = 0
			For k = 0 To nc - 1
				num = num + 1
				If dMx_R.x(i, j) = dMx_R.x(i, j) Then
					sum(0) = sum(0) + dMx_R.x(i, j) * M.dMx_R.x(i, j) - dMx_I.x(i, j) * M.dMx_I.x(i, j)
					sum(1) = sum(1) + dMx_R.x(i, j) * M.dMx_I.x(i, j) + dMx_I.x(i, j) * M.dMx_R.x(i, j)
				End If
			Next
			If num > 0 Then 
				newM.dMx_R.x(i, j) = nc * sum(0) / num
				newM.dMx_I.x(i, j) = nc * sum(1) / num
			Else
				newM.dMx_R.x(i, j) = 0/0
				newM.dMx_I.x(i, j) = 0/0
			End If
		Next
	Next
	Return methods(newM, "dot." & M.id)
End Sub

