﻿B4J=true
Group=Default Group
ModulesStructureVersion=1
Type=Class
Version=8.5
@EndOfDesignText@
Sub Class_Globals
	Private mCallBack As Object			'ignore
	Private dMx As dMatrix
	
	Public nDecimals As Int
End Sub

Public Sub Initialize(callBack As Object, numDecimals As Int)
	dMx.Initialize(callBack, "")
	mCallBack = callBack
	nDecimals = numDecimals
End Sub

'returns an Array of 2 dMatrix Objects, eigenVectors and eigenValues
'this method is limited to Symmetric matrices
'for square non-symmetric matrices (which can have complex eigenValues) look at the Schur decomposition
Public Sub eigenSymmetric(mat As dMatrix) As dMatrix()
	' KAISER,H.F. (1972) 
	' THE JK METHOD: A PROCEDURE FOR FINDING THE EIGENVALUES OF A REAL SYMMETRIC MATRIX', The Computer Journal, VOL.15, 271-273.
	' This B4X dMatrix version using dMatrix is adapted from https://www.freevbcode.com/ShowCode.asp?ID=9209
	Dim i, j, k, iter, p As Long
	Dim den, hold, num As Double
	Dim Sin1, Sin2, Cos1, Cos2, Tan1, Cot1 As Double
	Dim tmp, sum As Double
	Dim eps As Double = Power(10, -5 * nDecimals)
	p = mat.nc
	Dim vectors As dMatrix = mat.dValue("eigenVectors", p, p, 0)
	Dim values As dMatrix = vectors.copy.rename("eigenValues")
	Dim A As dMatrix = mat.copy
	For iter = 1 To 15
		For j = 0 To p - 2									'Orthogonalize pairs of columns in upper off diag
			For k = j + 1 To p - 1
				den = 0.0: 	num = 0.0
				For i = 0 To p - 1							'Perform single plane rotation
					num = num + 2 * A.x(i, j) * A.x(i, k)
					den = den + (A.x(i, j) + A.x(i, k)) * (A.x(i, j) - A.x(i, k))
				Next
				If Abs(num) < eps And den >= 0 Then Exit	'Skip rotation if a.x(i, j) is zero and correct ordering
				If Abs(num) <= Abs(den) Then				'Perform Rotation
					Tan1 = Abs(num) / Abs(den)
					Cos1 = 1 / Sqrt(1 + Tan1 * Tan1)
					Sin1 = Tan1 * Cos1
				Else
					Cot1 = Abs(den) / Abs(num)
					Sin1 = 1 / Sqrt(1 + Cot1 * Cot1)
					Cos1 = Cot1 * Sin1
				End If
				
				Cos2 = Sqrt((1 + Cos1) / 2)
				Sin2 = Sin1 / (2 * Cos2)
				If den < 0 Then
					tmp = Cos2
					Cos2 = Sin2
					Sin2 = tmp
				End If
				If num < 0 Then Sin2 = -Sin2
				For i = 0 To p - 1							'Rotate
					tmp = A.x(i, j)
					A.x(i, j) = tmp * Cos2 + A.x(i, k) * Sin2
					A.x(i, k) = -tmp * Sin2 + A.x(i, k) * Cos2
				Next
			Next
		Next
		sum = 0												'Test for convergence
		For i = 0 To p - 1
			For j = 0 To p - 1
				sum = sum + A.x(i, j) * A.x(i, j)
			Next
		Next
		If Abs(sum - hold) < eps And iter > 5 Then Exit
		hold = sum
	Next
	If iter = 16 Then Log("JK Iteration has not converged.")
	For j = 0 To p - 1										'Compute eigenvalues/eigenvectors
		For k = 0 To p - 1									'Compute eigenvalues
			values.x(j, j) = values.x(j, j) + A.x(k, j) * A.x(k, j)
		Next
		values.x(j, j) = Sqrt(values.x(j, j))
		For i = 0 To p - 1									
			If values.x(j, j) <= 0 Then
				vectors.x(i, j) = 0
			Else
				vectors.x(i, j) = A.x(i, j) / values.x(j, j)
			End If
		Next
	Next
	vectors = vectors.iterate(vectors, "reflect", Array(vectors.sumC, "row")).rename("eigenVectors")    	'reflect colummns with negative sum
	values.rename("eigenValues")
	Return Array As dMatrix(vectors, values)
End Sub

'returns an Array of 2 dMatrix Objects, Q (which is orthogonal) and R (which is triangular)
Public Sub QR(mat As dMatrix) As dMatrix()
	'This B4X dMatrix version using dMatrix Is adapted from JAMA
	Dim R As dMatrix = mat.copy
	Dim m As Int = mat.nr
	Dim n As Int = mat.nc
	Dim Rdiag As dMatrix = dMx.dIdentity("", n)
	Dim nrm, s As Double
	For k = 0 To n - 1
		nrm = 0
		For i = k To m - 1
			nrm = Sqrt(nrm * nrm + R.x(i, k) * R.x(i, k))
		Next
		If nrm <> 0.0 Then
			If R.x(k, k) < 0 Then nrm = -nrm
			For i = k To m - 1
				R.x(i, k) = R.x(i, k) / nrm
			Next
			R.x(k, k) = R.x(k, k) + 1.0
			For j = k+1 To n - 1
				s = 0
				For i = k To m - 1
					s = s + R.x(i, k) * R.x(i, j)
				Next
				s = -s / R.x(k, k)
				For i = k To m - 1
					R.x(i, j) = R.x(i, j) + s * R.x(i, k)
				Next
			Next
		End If
		Rdiag.x(k, k) = -nrm
	Next
	For i = 0 To m - 1
		For j = 0 To n - 1
			If i > j Then 
				R.x(i, j) = 0
			Else if i = j Then
				R.x(i, j) = Rdiag.x(i, j)
			End If
		Next
	Next
	Dim Q As dMatrix = R.dot(mat.inverse).transpose.rename("Q")
	R.rename("R")
	Return Array As dMatrix(Q, R)
End Sub

'returns an Array of 2 dMatrix Objects, Q (which is orthogonal) and triangular T (which is related to eigenValues)
Public Sub Schur(mat As dMatrix) As dMatrix()
	Dim eps As Double = Power(10, -2 * nDecimals)
	Dim QRResults() As dMatrix
	Dim Ak As dMatrix = mat.copy
	Dim Uk As dMatrix = dMx.dIdentity("", mat.nr)
	Dim lastOne As dMatrix = dMx.dZero("", mat.nr, mat.nr)
	Dim done As Boolean = True
	For k = 1 To 50
		QRResults = QR(Ak)
		Ak = QRResults(1).dot(QRResults(0))
		Uk = Uk.dot(QRResults(0))
		done = True
		For i = 0 To mat.nr - 1
			For j = 0 To mat.nr - 1
				If Abs(Ak.x(i, j) - lastOne.x(i, j)) > eps Then 
					done = False
					Exit
				End If
			Next
		Next
		If done Then 
			Exit
		End If
		lastOne = Ak.copy
	Next
	For i = 0 To mat.nr - 1
		For j = i To mat.nr - 1
			If Abs(Ak.x(i, j)) <= eps Then Ak.x(i, j) = 0
		Next
	Next
	Dim T As dMatrix = Ak.rename("T")
	Dim Q As dMatrix = Uk.rename("Q")
	Return Array As dMatrix(Q, T)
End Sub

'returns an Array of 3 dMatrix Objects, U (which is orthogonal), S (diagonal), and V (which is orthogonal)
Public Sub SVD (mat As dMatrix) As dMatrix()
'	Translated to B4X from JAMA
'   For an m-by-n matrix A the singular value decomposition is
'   an m-by-n orthogonal matrix U, an n-by-n diagonal matrix S, and
'   an n-by-n orthogonal matrix V so that A = U * S * V'.
'
'   The singular values, sigma[k] = S[k][k], are ordered so that
'   sigma[0] >= sigma[1] >= ... >= sigma[n-1].
'
'   The singular value decompostion (svd) always exists, so the constructor will
'   never fail.  The matrix condition number and the effective numerical
'   rank can be computed from this decomposition.

	Private  U(,), V(,) As Double
	Private  s() As Double
	Private  m, n As Int
	Dim uPart, sPart, vPart As dMatrix
	
	If mat.nr < mat.nc Then 
		Dim results() As dMatrix = SVD(mat.transpose)
		uPart = results(2).rename("U")
		vPart = results(0).rename("V")
		Return Array As dMatrix(uPart, results(1), vPart)
	Else
		Dim A(,) As Double = mat.copy.x
	End If

	m = mat.nr: n = mat.nc
	Dim nu As Int = Min(m,n)
	Dim s(Min(m + 1,n)) As Double
	Dim U(m, nu) As Double
	Dim V(n, n) As Double
	Dim e(n) As Double
	Dim work(m) As Double
	
	'Reduce A to bidiagonal form, storing the diagonal elements in s and the super-diagonal elements in e.
	Dim nct As Int = Min(m - 1, n)
	Dim nrt As Int = Max(0, Min(n - 2, m))
	For k = 0 To Max(nct, nrt) - 1
		If k < nct Then
			'Compute the transformation for the k-th column and place the k-th diagonal in s[k].
			'Compute 2-norm of k-th column without under / overflow.
			s(k) = 0
			For i = k To m - 1
				s(k) = Sqrt(s(k) * s(k) + A(i, k) * A(i, k))
			Next
			If s(k) <> 0 Then
				If A(k, k) < 0 Then s(k) = -s(k)
				For i = k To m - 1
					A(i, k) = A(i, k) / s(k)
				Next
				A(k, k) = A(k, k) + 1
			End If
			s(k) = -s(k)
		End If
		For j = k + 1 To n - 1
			If k < nct And s(k) <> 0  Then
				'Apply the transformation.
				Dim t As Double = 0
				For i = k To m - 1
					t = t + A(i, k) * A(i, j)
				Next
				t = -t / A(k, k)
				For i = k To m - 1
					A(i, j) = A(i, j) + t * A(i, k)
				Next
			End If
	
			'Place the k-th row of A into e for the subsequent calculation of the row transformation.
			e(j) = A(k, j)
		Next
		If k < nct Then
			'Place the transformation in U for subsequent back multiplication.
			For i = k To m - 1
				U(i, k) = A(i, k)
			Next
		End If
		If k < nrt Then 
			'Compute the k-th row transformation and place the k-th super-diagonal in e[k].
			'Compute 2-norm without under / overflow.
			e(k) = 0
			For i = k + 1 To n - 1
				e(k) = Sqrt(e(k) * e(k) + e(i) * e(i))
			Next
			If e(k) <> 0 Then
				If (e(k + 1) < 0) Then e(k) = -e(k)
				For i = k + 1 To n - 1
					e(i) = e(i) / e(k)
				Next
				e(k + 1) = e(k + 1) + 1
			End If
			e(k) = -e(k)
			If k + 1 < m And e(k) <> 0 Then
				'Apply the transformation.
				For i = k + 1 To m - 1
					work(i) = 0
				Next
				For j = k + 1 To n - 1
					For i = k + 1 To m - 1
						work(i) = work(i) + e(j) * A(i, j)
					Next
				Next
				For j = k + 1 To n - 1
					Dim t As Double = -e(j) / e(k + 1)
					For i = k + 1 To m - 1
						A(i, j) = A(i, j) + t * work(i)
					Next
				Next
			End If
			'Place the transformation in V for subsequent back multiplication.
			For i = k + 1 To n - 1
				V(i, k) = e(i)
			Next
		End If
	Next

	'Set up the final bidiagonal matrix or order p.
	Dim p As Int = Min(n,m + 1)
	If nct < n Then s(nct) = A(nct, nct)
	If m < p Then s(p - 1) = 0
	If nrt + 1 < p Then e(nrt) = A(nrt, p - 1)
	e(p - 1) = 0

	'Generate U.
	For j = nct To nu - 1
		For i = 0 To m - 1
			U(i, j) = 0
		Next
		U(j, j) = 1
	Next
	For k = nct - 1 To 0 Step -1
		If s(k) <> 0 Then
			For j = k + 1 To nu - 1
				Dim t As Double = 0
				For i = k To m - 1
					t = t + U(i, k) * U(i, j)
				Next
				t = -t / U(k, k)
				For i = k To m - 1
					U(i, j) = U(i, j) + t * U(i, k)
				Next
			Next
			For i = k To m - 1
				U(i, k) = -U(i, k)
			Next
			U(k, k) = 1 + U(k, k)
			For i = 0 To k - 2
				U(i, k) = 0
			Next
		Else
			For i = 0 To m - 1
				U(i, k) = 0
			Next
			U(k, k) = 1
		End If
	Next
	
	'Generate V.
	For k = n - 1 To 0 Step -1
		If k < nrt And e(k) <> 0 Then
			For j = k + 1 To nu - 1
				Dim t As Double = 0
				For i = k + 1 To n - 1
					t = t + V(i, k) * V(i, j)
				Next
				t = -t / V(k + 1, k)
				For i = k + 1 To n - 1
					V(i, j) = V(i, j) + t * V(i, k)
				Next
			Next
		End If
		For i = 0 To n - 1
			V(i, k) = 0
		Next
		V(k, k) = 1
	Next
	
	'Main iteration loop for the singular values.
	Dim pp As Int = p - 1
	Dim iter As Int = 0
	Dim eps As Double = Power(10, -5 * nDecimals)
	Dim tiny As Double = Power(2.0,-966.0)
	Do While p > 0
        If iter > 30 Then 
			Log("No convergence")
			Return Array As Object()
		End If
		Dim kase As Int
		'This section of the program inspects for
		'negligible elements in the s and e arrays.  On
		'completion the nature of variables and k are set as follows.
	
		'kase = 1     if s(p) and e[k-1] are negligible and k<p
		'kase = 2     if s(k) is negligible and k<p
		'kase = 3     if e[k-1] is negligible, k<p, and s(k), ..., s(p) are not negligible (qr step).
		'kase = 4     if e(p-1) is negligible (convergence).
	
		For k = p - 2 To -1 Step -1
			If k = -1 Then Exit
			If Abs(e(k)) <= tiny + eps * (Abs(s(k) ) + Abs(s(k + 1))) Then
				e(k) = 0
				Exit
			End If
		Next
		If k = p - 2 Then
			kase = 4
		Else
			Dim ks As Int
			For ks = p - 1 To k Step -1
				If ks = k Then Exit
				Dim t As Double
				If ks <> p Then t = Abs(e(ks))
				If ks <> k + 1 Then t = t + Abs(e(ks - 1))
				If Abs(s(ks)) <= tiny + eps * t  Then
					s(ks) = 0
					Exit
				End If
			Next
			If ks = k Then
				kase = 3
			Else If ks = p - 1 Then
				kase = 1
			Else
				kase = 2
				k = ks
			End If
		End If
		k = k + 1
	
		'Perform the task indicated by kase.
		Select kase
			Case 1: 						'Deflate negligible s(p)
				Dim f As Double = e(p - 2)
				e(p - 2) = 0
				For j = p - 2 To k Step -1
					Dim t As Double = Sqrt(s(j) * s(j) + f * f)
					Dim cs As Double = s(j) / t
					Dim sn As Double = f / t
					s(j) = t
					If j <> k Then
						f = -sn * e(j - 1)
						e(j - 1) = cs * e(j - 1)
					End If
					For i = 0 To n - 1
						t = cs * V(i, j) + sn * V(i, p - 1)
						V(i, p - 1) = -sn * V(i, j) + cs * V(i, p - 1)
						V(i, j) = t
					Next
				Next
			Case 2:			 			'Split at negligible s(k)
				Dim f As Double = e(k - 1)
				e(k - 1) = 0
				For j = k To p - 1
					Dim t As Double = Sqrt(s(j) * s(j) + f * f)
					Dim cs As Double = s(j) / t
					Dim sn As Double = f / t
					s(j) = t
					f = -sn * e(j)
					e(j) = cs * e(j)
					For i = 0 To m - 1
						t = cs * U(i, j) + sn * U(i, k - 1)
						U(i, k - 1) = -sn * U(i, j) + cs * U(i, k - 1)
						U(i, j) = t
					Next
				Next
			Case 3:		 							'Perform one qr step
				'Calculate the shift.
				Dim scale As Double = Max(Max(Max(Max(Abs(s(p - 1)),Abs(s(p - 2))),Abs(e(p - 2))), Abs(s(k))),Abs(e(k)))
				Dim sp As Double = s(p - 1) / scale
				Dim spm1 As Double = s(p - 2) / scale
				Dim epm1 As Double = e(p - 2) / scale
				Dim sk As Double = s(k) / scale
				Dim ek As Double = e(k) / scale
				Dim b As Double = ((spm1 + sp) * (spm1 - sp) + epm1 * epm1) / 2.0
				Dim c As Double = (sp * epm1) * (sp * epm1)
				Dim shift As Double = 0
				If b <> 0 Or c <> 0 Then
					shift = Sqrt(b * b + c)
					If b < 0 Then shift = -shift
					shift = c / (b + shift)
				End If
				Dim f As Double = (sk + sp) * (sk - sp) + shift
				Dim g As Double = sk * ek
					
				'Chase zeros.
				For j = k To p - 2
					Dim t As Double = Sqrt(f * f + g * g)
					Dim cs As Double = f / t
					Dim sn As Double = g / t
					If j <> k Then e(j - 1) = t
					f = cs * s(j) + sn * e(j)
					e(j) = cs * e(j) - sn * s(j)
					g = sn * s(j + 1)
					s(j + 1) = cs * s(j + 1)
					For i = 0 To n - 1
						t = cs * V(i, j) + sn * V(i, j + 1)
						V(i, j + 1) = -sn * V(i, j) + cs * V(i, j + 1)
						V(i, j) = t
					Next
					t = Sqrt(f * f + g * g)
					cs = f / t
					sn = g / t
					s(j) = t
					f = cs * e(j) + sn * s(j + 1)
					s(j + 1) = -sn * e(j) + cs * s(j + 1)
					g = sn * e(j + 1)
					e(j + 1) = cs * e(j + 1)
					If j < m - 1 Then
						For i = 0 To m - 1
							t = cs * U(i, j) + sn * U(i, j + 1)
							U(i, j + 1) = -sn * U(i, j) + cs * U(i, j + 1)
							U(i, j) = t
						Next
					End If
				Next
				e(p - 2) = f
				iter = iter + 1
			Case 4: 				'Convergence.
				'Make the singular values positive.
				If s(k) <= 0 Then
					If s(k) < 0 Then 
						s(k) = -s(k)
					Else
						s(k) = 0
					End If
					For i = 0 To pp
						V(i, k) = -V(i, k)
					Next
				End If
					
				'Order the singular values.
				Do While k < pp
					If s(k) >= s(k + 1) Then Exit
					Dim t As Double = s(k)
					s(k) = s(k + 1)
					s(k + 1) = t
					If k < n - 1 Then
						For i = 0 To n - 1
							t = V(i, k + 1) 
							V(i, k + 1) = V(i, k)
							V(i, k) = t
						Next
					End If
					If k < m - 1 Then
						For i = 0 To m - 1
							t = U(i, k + 1)
							U(i, k + 1) = U(i, k) 
							U(i, k) = t
						Next
					End If
					k = k + 1
				Loop
				iter = 0
				p = p - 1
		End Select
	Loop

	sPart = dMx.dValue("S", Min(m + 1, n), Min(m + 1, n), 0)
	For i = 0 To sPart.nr - 1
		sPart.x(i, i) = s(i)
	Next
	uPart = dMx.dValue("U", m, nu, 0)
	uPart.x = U
	
	vPart = dMx.dValue("V", n, n, 0)
	vPart.x = V
	Return Array As dMatrix(uPart, sPart, vPart)
End Sub

